@@ -1,4 +1,5 @@
 from scipy.stats import invwishart, multivariate_normal
+from scipy.special import digamma, loggamma
 from itertools import repeat
 import numpy as np
 
@@ -72,20 +73,50 @@ def mixture_density(x, w, mu, sigma, u):
     return ret
 
 
-def cluster(x, w, mu, sigma, u):
+def cluster(x, w, mu, sigma, u=None):
     k = len(w)
-    ret = []
+    assign_prob = []
     for j in range(k):
-        ret.append(multivariate_normal.pdf(x,
-                                           mu[j],
-                                           sigma[j],
-                                           1))
-    ret = np.array(ret).T
-    mask = (np.array(list(repeat(u, k))) <
-            np.array(list(repeat(w, len(u)))).transpose())
+        assign_prob.append(multivariate_normal.pdf(x,
+                                                   mu[j],
+                                                   sigma[j],
+                                                   1))
+    assign_prob = np.array(assign_prob).T
+    if u is not None:
+        mask = (np.array(list(repeat(u, k))) <
+                np.array(list(repeat(w, len(u)))).transpose())
 
-    weights = (mask / mask.sum(0)).sum(1) / len(u)
-    ret = ret * weights
-    grp = np.argmax(ret, axis=1)
+        w = (mask / mask.sum(0)).sum(1) / len(u)
+    assign_prob = assign_prob * w
+    grp = np.argmax(assign_prob, axis=1)
+    uncertainty = 1 - assign_prob[range(len(x)), grp]
     u_grp, ret = np.unique(grp, return_inverse=True)
-    return ret, weights[u_grp], mu[u_grp], sigma[u_grp]
+    return ret, uncertainty, w[u_grp], mu[u_grp], sigma[u_grp]
+
+
+def log_wishart_normalization_term(precision, scale):
+    dim = precision.shape[0]
+    res = np.log(np.linalg.norm(precision)) * (-scale / 2)
+    inverse_term = np.log(2) * (scale * dim / 2)
+    inverse_term += np.log(np.pi) * (dim * (dim - 1) / 4)
+    for i in range(dim):
+        inverse_term += loggamma((scale - i) / 2)
+    res -= inverse_term
+    return res
+
+
+def e_log_norm_wishart(precision, scale):
+    dim = precision.shape[0]
+    res = np.log(np.linalg.norm(precision))
+    res += dim * np.log(2)
+    for i in range(dim):
+        res += digamma((scale - i) / 2)
+    return res
+
+
+def entropy_wishart(precision, scale):
+    dim = precision.shape[0]
+    res = -log_wishart_normalization_term(precision, scale)
+    res -= (scale - dim - 1) / 2 * e_log_norm_wishart(precision, scale)
+    res += scale * dim / 2
+    return res
